import logging
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

import os
import time
import boto3
from sagemaker.session import Session
from sagemaker.estimator import Estimator, Transformer
from sagemaker import get_execution_role

logger = logging.getLogger(None)


class CreateTrainingJob(BaseOperator):
    template_fields = ('config', 'xcom_tasks')

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateTrainingJob, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            May need to change input to be AthenaFeatureQuery

            https://sagemaker.readthedocs.io/en/stable/api/prep_data/\
                feature_store.html#sagemaker.feature_store.inputs.DataCatalogConfig
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_CLIENT = BOTO_SESSION.client('sagemaker')
        SM_SESSION = Session(default_bucket=self.config['bucket'], boto_session=BOTO_SESSION)
        # get execution role
        if self.config.get('role', None) in ['None', None, '']:
            print("Running on sagemaker friendly context")
            ROLE = get_execution_role(sagemaker_session=SM_SESSION)
        else:
            print("Running else where")
            IAM_CLIENT = BOTO_SESSION.client('iam')
            ROLE = IAM_CLIENT.get_role(RoleName=self.config['sagemaker_role_name'])['Role']['Arn']

        input_data = self.config.get('inputs', None)

        try:
            if input_data is None:
                input_data = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['input_data']['task_id'],
                    key=self.xcom_tasks['input_data']['key']
                )
                print("input_data = {}".format(input_data))
                assert input_data is not None, "No training job name found. Either hardcode in config or pass from previous job"
            print("creating estimator")
            estimator = Estimator(
                image_uri=self.config['image_uri'],
                role=ROLE,
                train_instance_type=self.config.get('train_instance_type', 'ml.t3.medium'),
                train_instance_count=self.config.get('train_instance_count', 1),
                train_max_run=self.config.get('train_max_run', None),
                output_path=self.config['output_path'],
                base_job_name=self.config['base_job_name'],
                sagemaker_session=SM_SESSION,
                hyperparameters=self.config.get('hyperparameters', None),
                tags=self.config.get('tags', None),
                enable_sagemaker_metrics=self.config.get('enable_sagemaker_metrics',False),
                metric_definitions=self.config.get('metric_definitions',{})
            )

            print("fitting data")
            # add test data to s3 s3://bucket/bernard/data
            estimator.fit(
                inputs = input_data,
                wait=self.config.get('wait', False)
            )

            context['task_instance'].xcom_push(
                key='training_job_name',
                value=estimator.latest_training_job.name
            )
            context['task_instance'].xcom_push(
                key='model_name',
                value=estimator.latest_training_job.name
            )
        except Exception as exception:
            raise exception

class CreateModel(BaseOperator):
    template_fields = ('config', 'xcom_tasks', )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateModel, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            Create model through usual boto3 handling

            model_name = training_job_name + '-mod'

            info = sm.describe_training_job(TrainingJobName=training_job_name)
            model_data = info['ModelArtifacts']['S3ModelArtifacts']
            print(model_data)

            primary_container = {
                'Image': container,
                'ModelDataUrl': model_data
            }

            create_model_response = sm.create_model(
                ModelName = model_name,
                ExecutionRoleArn = role,
                PrimaryContainer = primary_container)

            print(create_model_response['ModelArn'])
        """
        import os
        import boto3
        from botocore.exceptions import ClientError as botocore_clienterror
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_CLIENT = BOTO_SESSION.client('sagemaker')

        # context = context_dict + op_kwargs_dict
        training_job_name = self.config.get('training_job_name', None) # pass from training job
        container_image = self.config.get('container_image', None)

        try:
            if training_job_name is None:
                training_job_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['training_job_name']['task_id'],
                    key=self.xcom_tasks['training_job_name']['key']
                )
                print("training_job_name = {}".format(training_job_name))
                assert training_job_name is not None, "No training job name found. Either hardcode in config or pass from previous job"

            # get model_name
            model_name = self.config.get('model_name', training_job_name)

            # retrieve training job config
            training_job_config = SM_CLIENT.describe_training_job(
                TrainingJobName=training_job_name
            )

            if container_image is None:
                container_image=training_job_config['AlgorithmSpecification']['TrainingImage']

            primary_container = {
                'Image': container_image,
                'Mode': self.config.get('primary_container', {}).get('mode','SingleModel'),
                'ModelDataUrl': training_job_config['ModelArtifacts']['S3ModelArtifacts'],
                'Environment': self.config.get('primary_container', {}).get('environment', {})
            }

            model_tags = self.config.get('tags', [])
            model_tags.append({
                'Key': 'status',
                'Value': 'active'
            })
            try:
                _ = SM_CLIENT.create_model(
                    ModelName=model_name,
                    ExecutionRoleArn=training_job_config['RoleArn'],
                    PrimaryContainer=primary_container,
                    Tags=model_tags
                )

            except botocore_clienterror as exception:
                if exception.response['Error']['Code'] == 'ValidationException':
                    # handle no
                    if 'Cannot create already existing model' in exception.response['Error']['Message']:
                        print((
                            "This Model already exists, "
                            "with name = {}. Using the already created model."
                            .format(model_name)
                        ))
                    else:
                        raise botocore_clienterror(
                            error_response=exception.operation_name,
                            operation_name=exception.response
                        )

            context['task_instance'].xcom_push(
                key='model_name',
                value=model_name
            )
        except Exception as err:
            raise

class CreateTransformJob(BaseOperator):
    template_fields = ('config', 'xcom_tasks', )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateTransformJob, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            Creates a transform job

            https://sagemaker.readthedocs.io/en/stable/api/inference/transformer.html#sagemaker.transformer.Transformer
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_CLIENT = BOTO_SESSION.client('sagemaker')
        SM_SESSION = Session(default_bucket=self.config['bucket'], boto_session=BOTO_SESSION)
        # get execution role
        if self.config.get('role', None) in ['None', None, '']:
            print("Running on sagemaker friendly context")
            ROLE = get_execution_role(sagemaker_session=SM_SESSION)
        else:
            print("Running else where")
            IAM_CLIENT = BOTO_SESSION.client('iam')
            ROLE = IAM_CLIENT.get_role(RoleName=self.config['sagemaker_role_name'])['Role']['Arn']

        model_name = self.config.get('model_name', None)
        data = self.config.get('data', None)

        try:
            if model_name is None:
                # assume model_name is passed
                model_name = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['model_name']['task_id'],
                    key=self.xcom_tasks['model_name']['key']
                )
                print("model_name = {}".format(model_name))
                logger.info("successfully retrieved model_name = {}".format(model_name))
                assert model_name is not None, "No model name found. Either hardcode in config or pass from previous job"
            
            if data is None:
                # assume model_name is passed
                data = context['task_instance'].xcom_pull(task_ids=self.xcom_tasks['data']['task_id'], key=self.xcom_tasks['data']['key'])
                print("data = {}".format(data))
                logger.info("successfully retrieved data path = {}".format(data))
                assert data is not None, "No s3 data path found. Either hardcode in config or pass from previous job"       

            transformer = Transformer(
                model_name=model_name,
                instance_count=self.config['instance_count'],
                instance_type=self.config['instance_type'],
                strategy=self.config.get('strategy', None),
                assemble_with=self.config.get('split_type', None),
                output_path=self.config['output_path'],
                output_kms_key=self.config.get('output_kms_key', None),
                accept=self.config.get('content_type', None),
                max_concurrent_transforms=self.config.get('max_concurrent_transforms', None),
                max_payload=self.config.get('max_payload', None),
                tags=self.config.get('tags', None),
                env=self.config.get('env', None),
                base_transform_job_name=self.config['base_transform_job_name'],
                sagemaker_session=SM_SESSION
            )
            transformer.transform(
                data=data,
                data_type=self.config.get('data_type','S3Prefix'),
                split_type=self.config.get('split_type','Line'),
                compression_type=self.config.get('compression_type', None),
                content_type=self.config.get('content_type','text/csv'),
                input_filter=self.config.get('input_filter',None),
                join_source=self.config.get('join_source',None),
                output_filter=self.config.get('output_filter', None),
                wait=self.config.get('wait', True)
            )

            context['task_instance'].xcom_push(
                key='transform_job_name',
                value=transformer.latest_transform_job.name
            )
        except Exception as err:
            raise

class CreateEndpointConfig(BaseOperator):
    template_fields = ('config', 'xcom_tasks', )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(CreateEndpointConfig, self).__init__(*args, **kwargs)
        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """get model configuration"""
        import os
        import boto3
        from botocore.exceptions import ClientError as botocore_clienterror
        from utils.sagemaker import \
            build_resource_tags, name_from_base
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_CLIENT = BOTO_SESSION.client('sagemaker')
        up_stream = op_args[0]
        model_name = context.get('model_name', None)
        try:
            if model_name is None:
                # assume model_name is passed
                model_name = context['task_instance'].xcom_pull(task_ids=self.xcom_task_id, key=self.xcom_task_id_key)
                print("model_name = {}".format(model_name))
                assert model_name is not None, "No model name found. Either hardcode in config or pass from previous job"
                response = client.create_endpoint_config(
                    EndpointConfigName=context.get('endpoint_configuration', {})['endpoint_config_name'],
                    ProductionVariants=context.get('endpoint_configuration', {})['production_variants'],
                    DataCaptureConfig=context.get('endpoint_configuration', {})['data_capture_config'],
                    Tags=context.get('endpoint_configuration', {})['tags']
                )
            return model_name
        except Exception as err:
            raise



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    GLUE TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import json
import logging
from datetime import datetime, timedelta

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class StartQuery(BaseOperator):
    """
    Queries CloudWatch Logs. CloudWatch Logs Insights supports a query language
    you can use to perform queries on your log groups. Each query can include one
    or more query commands.
    Query Syntax: https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/CWL_QuerySyntax.html

    Queries time out after 15 minutes of execution.
    If your queries are timing out, reduce the time range being searched
    or partition your query into a number of queries.
    """
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(StartQuery, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :config.job_name: str. the name of the glue job to run.
            :config.job_arguments: dict. the key value arguments send to the job.
            :xcom_push: str. glue_job_run_id.
        """
        LOGS_CLIENT = boto3.client('logs')

        query = self.config.get('query', 'fields @timestamp, @message | sort @timestamp desc')
        log_groups = self.config.get('log_groups', None)
        hour_delta = self.config.get('hour_delta', 5)
        limit = self.config.get('query', 100)

        try:
            # Start query and gather query id
            query_id = LOGS_CLIENT.start_query(
                logGroupNames=log_groups,
                # filter by current time to specified time delta
                startTime=int((datetime.today() - timedelta(hours=hour_delta)).timestamp()),
                endTime=int(datetime.now().timestamp()),
                queryString=query,
                limit=limit
            )['queryId']

            context['task_instance'].xcom_push(
                key='query_id',
                value=query_id
            )

        except Exception as err:
            raise err


class LogToS3(BaseOperator):
    """
    Returns the results from the specified query made from a log group
    using CloudWatch Logs Insights. The data returned is saved as
    json in S3 in the specified bucket and key.
    """
    template_fields = ('config', 'xcom_tasks',)

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(LogToS3, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
            :config.job_name: str. the name of the glue job to run.
            :config.job_arguments: dict. the key value arguments send to the job.
            :xcom_push: str. glue_job_run_id.
        """
        LOGS_CLIENT = boto3.client('logs')
        S3_CLIENT = boto3.client('s3')

        s3_bucket = self.config.get('s3_bucket', None)
        s3_key = self.config.get('s3_key', None)

        if s3_bucket is None or s3_key is None:
            raise ValueError('S3 location is not specified')

        query_id = self.config.get('query_id', None)
        task_ids = self.xcom_tasks['query_id']['task_id']

        if query_id is None:
            query_id = context['task_instance'].xcom_pull(
                task_ids=task_ids,
                key=self.xcom_tasks['query_id']['key']
            )

        try:
            cloud_watch_logs = LOGS_CLIENT.get_query_results(
                queryId=query_id
            )

            S3_CLIENT.put_object(
                Body=json.dumps(cloud_watch_logs),
                Bucket=s3_bucket,
                Key=f'{s3_key}/{datetime.now()}.json'
            )
        except Exception as err:
            raise err



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""SDC Custom S3 Operators"""
"""SDC Custom S3 Operators"""
import os
import logging
from urllib.parse import urlparse
import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)

def parse_s3_url(url):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3 scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != "s3":
        raise ValueError("Expecting 's3' scheme, got: {} in {}.".format(parsed_url.scheme, url))
    return parsed_url.netloc, parsed_url.path.lstrip("/")

class CopyBetweenS3Locations(BaseOperator):
    template_fields = ('dst_path', 'src_path', 'keep_filename', 'xcom_tasks' )

    @apply_defaults
    def __init__(self, dst_path, src_path=None, keep_filename=False, xcom_tasks=None, *args, **kwargs):
        super(CopyBetweenS3Locations, self).__init__(*args, **kwargs)
        self.dst_path = dst_path
        self.src_path = src_path
        self.xcom_tasks = xcom_tasks
        self.keep_filename = keep_filename

    def execute(self, context):
        """ Copy s3 object between two buckets
        """
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_RESOURCE = BOTO_SESSION.resource('s3')

        try:
            if self.src_path is None:
                self.src_path = context['task_instance'].xcom_pull(
                    task_ids=self.xcom_tasks['src_path']['task_id'],
                    key=self.xcom_tasks['src_path']['key']
                )
                assert self.src_path is not None, "No source path found. Either hardcode in config or pass from previous job"

            src_bucket, src_filepath = parse_s3_url(self.src_path)
            dst_bucket, dst_filepath = parse_s3_url(self.dst_path)

            if self.keep_filename:
                # assumes destination path is a folder path
                # keeps src filename
                src_filename = src_filepath.split('/')[-1]
                logger.info("Keep source filename = {}".format(src_filename))
                # assumes dst_path given without filename
                dst_filepath = os.path.join(dst_filepath, src_filename)
                # assumes dst_path is given without filename
                self.dst_path = os.path.join(self.dst_path, src_filename)

            copy_source = {
                'Bucket': src_bucket,
                'Key': src_filepath
            }
            bucket = SM_RESOURCE.Bucket(dst_bucket)
            bucket.copy(copy_source, dst_filepath)

            # push response up
            context['task_instance'].xcom_push(
                key='destination_location',
                value=self.dst_path
            )

        except Exception as exception:
            raise exception



import logging

import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class MonitorGlueJob(BaseSensorOperator):
    """
    Retrieves the metadata for a given job run and checks the run status.
    """
    template_fields = ('poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(self, poke_interval=60, timeout=60 * 60 * 24 * 7, soft_fail=False, mode='poke',
                 xcom_tasks=None, *args, **kwargs):
        super(MonitorGlueJob, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors Glue Job Run Status"""

        CLIENT = boto3.client('glue')

        task_ids = self.xcom_tasks['job_name']['task_id']

        job_name = context['task_instance'].xcom_pull(
            task_ids=task_ids,
            key=self.xcom_tasks['job_name']['key']
        )

        job_run_id = context['task_instance'].xcom_pull(
            task_ids=task_ids,
            key=self.xcom_tasks['job_run_id']['key']
        )

        try:
            response = CLIENT.get_job_run(
                JobName=job_name,
                RunId=job_run_id,
            )

            state = response['JobRun']['JobRunState']
            if state == 'SUCCEEDED':
                # success
                return True
            elif state in ['STARTING', 'RUNNING', 'STOPPING']:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=state, REASON=response['JobRun']['ErrorMessage']
                    )
                )

        except Exception as err:
            raise


class MonitorGlueCrawler(BaseSensorOperator):
    """
    Retrieves the metadata for a given crawler run and checks the run status.
    """
    template_fields = ('poke_interval', 'timeout', 'soft_fail', 'mode', 'xcom_tasks')

    @apply_defaults
    def __init__(self, poke_interval=60, timeout=60 * 60 * 24 * 7, soft_fail=False, mode='poke',
                 xcom_tasks=None, *args, **kwargs):
        super(MonitorGlueCrawler, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs
        )
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors Glue Crawler Run Status"""

        CLIENT = boto3.client('glue')

        task_ids = self.xcom_tasks['crawler_name']['task_id']

        crawler_name = context['task_instance'].xcom_pull(
            task_ids=task_ids,
            key=self.xcom_tasks['crawler_name']['key']
        )

        try:
            response = CLIENT.get_crawler(
                Name=crawler_name
            )

            state = response['Crawler']['State']
            if state == 'READY':
                # success
                return True
            elif state in ['RUNNING', 'STOPPING']:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=state, REASON=state
                    )
                )

        except Exception as err:
            raise
